@@ -20,13 +20,13 @@
 #
 #
 
-from casadi import MX, substitute, Function, vcat, depends_on, vertcat, jacobian, veccat, jtimes, hcat, linspace, DM, constpow, mtimes, vvcat, low, floor, hcat, horzcat, DM
+from casadi import MX, substitute, Function, vcat, depends_on, vertcat, jacobian, veccat, jtimes, hcat, linspace, DM, constpow, mtimes, vvcat, low, floor, hcat, horzcat, DM, is_equal
 from .freetime import FreeTime
 from .direct_method import DirectMethod
 from .multiple_shooting import MultipleShooting
 from .single_shooting import SingleShooting
 from collections import defaultdict
-from .casadi_helpers import DM2numpy, get_meta, merge_meta
+from .casadi_helpers import DM2numpy, get_meta, merge_meta, HashDict, HashDefaultDict, HashOrderedDict
 from contextlib import contextmanager
 from collections import OrderedDict
 
@@ -71,31 +71,52 @@ class Stage:
         self.parameters = defaultdict(list)
         self.variables = defaultdict(list)
 
-        self.master = parent.master if parent else None
+        self._master = parent.master if parent else None
         self.parent = parent
 
-        self._param_vals = dict()
-        self._state_der = dict()
+        self._param_vals = HashDict()
+        self._state_der = HashDict()
+        self._state_next = HashDict()
         self._alg = []
         self._constraints = defaultdict(list)
         self._objective = 0
-        self._initial = OrderedDict()
+        self._initial = HashOrderedDict()
 
-        self._placeholders = dict()
-        self._placeholder_callbacks = dict()
-        self._offsets = dict()
-        self._inf_inert = OrderedDict()
-        self._inf_der = OrderedDict()
-        self.t = MX.sym('t')
+        self._placeholders = HashDict()
+        self._placeholder_callbacks = HashDict()
+        self._offsets = HashDict()
+        self._inf_inert = HashOrderedDict()
+        self._inf_der = HashOrderedDict()
+        self._t = MX.sym('t')
         self._stages = []
         self._method = DirectMethod()
         self._t0 = t0
         self._T = T
-        self.T = self._create_placeholder_expr(0, 'T')
-        self.t0 = self._create_placeholder_expr(0, 't0')
-        self.tf = self.T + self.t0
+        self._public_T = self._create_placeholder_expr(0, 'T')
+        self._public_t0 = self._create_placeholder_expr(0, 't0')
+        self._tf = self.T + self.t0
         if not clone:
             self._check_free_time()
+    
+    @property
+    def master(self):
+        return self._master
+
+    @property
+    def t(self):
+        return self._t
+
+    @property
+    def T(self):
+        return self._public_T
+
+    @property
+    def t0(self):
+        return self._public_t0
+
+    @property
+    def tf(self):
+        return self._tf
 
     def _check_free_time(self):
         if isinstance(self._t0, FreeTime):
@@ -169,8 +190,9 @@ class Stage:
         >>> ocp.set_der(x, -x)
         >>> ocp.set_initial(x, sin(ocp.t)) # Optional: give initial guess
         """
+        import numpy
         # Create a placeholder symbol with a dummy name (see #25)
-        x = MX.sym("x", n_rows, n_cols)
+        x = MX.sym("x"+str(int(numpy.random.rand()*10000)), n_rows, n_cols)
         if quad:
             self.qstates.append(x)
         else:
@@ -261,14 +283,15 @@ class Stage:
 
     def set_value(self, parameter, value):
         if self.master is not None and self.master.is_transcribed:
-            self._method.set_value(self, self.master.opti, parameter, value)            
+            self._method.set_value(self, self.master._method, parameter, value)            
         else:
             self._param_vals[parameter] = value
 
     def set_initial(self, var, value):
+        assert "opti" not in str(var)
         self._initial[var] = value
         if self.master is not None and self.master.is_transcribed:
-            self._method.set_initial(self, self.master.opti, self._initial)
+            self._method.set_initial(self, self.master._method, self._initial)
 
     def set_der(self, state, der):
         """Assign a right-hand side to a state derivative
@@ -290,8 +313,32 @@ class Stage:
         >>> ocp.set_der(x, -x)
         """
         self._set_transcribed(False)
+        assert not self._state_next
         self._state_der[state] = der
 
+    def set_next(self, state, next):
+        """Assign an update rule for a discrete state
+
+        Parameters
+        ----------
+        state : `~casadi.MX`
+            A CasADi symbol created with :obj:`~rockit.stage.Stage.state`.
+        next : `~casadi.MX`
+            A CasADi symbolic expression of the same size as `state`
+
+        Examples
+        --------
+
+        Defining the first-order difference equation :  :math:`x^{+} = -x`
+        
+        >>> ocp = Ocp()
+        >>> x = ocp.state()
+        >>> ocp.set_next(x, -x)
+        """
+        self._set_transcribed(False)
+        self._state_next[state] = next
+        assert not self._state_der
+
     def add_alg(self, constr):
         self._set_transcribed(False)
         self._alg.append(constr)
@@ -300,7 +347,7 @@ class Stage:
         if depends_on(expr, self.u):
             raise Exception("Dependency on controls not supported yet for stage.der")
         ode = self._ode()
-        return jtimes(expr, self.x, ode(x=self.x, u=self.u, z=self.z, p=self.p, t=self.t)["ode"])
+        return jtimes(expr, self.x, ode(x=self.x, u=self.u, z=self.z, p=vertcat(self.p, self.v), t=self.t)["ode"])
 
     def integral(self, expr, grid='inf'):
         """Compute an integral or a sum
@@ -388,7 +435,14 @@ class Stage:
         """
         return self.offset(expr, -1)
 
-    def subject_to(self, constr, grid=None):
+    def clear_constraints(self):
+        """
+        Remove any previously declared constraints from the problem
+        """
+        self._set_transcribed(False)
+        self._constraints = defaultdict(list)
+
+    def subject_to(self, constr, grid=None,include_first=True,include_last=True,meta=None):
         """Adds a constraint to the problem
 
         Parameters
@@ -421,6 +475,7 @@ class Stage:
         >>> ocp.subject_to( ocp.at_tf(x) == 0)  # boundary constraint
         """
         self._set_transcribed(False)
+        #import ipdb; ipdb.set_trace()
         if grid is None:
             grid = 'control' if self.is_signal(constr) else 'point'
         if grid not in ['point', 'control', 'inf', 'integrator', 'integrator_roots']:
@@ -432,8 +487,8 @@ class Stage:
             if grid != 'point': 
                 raise Exception("Expected signal expression since grid '" + grid + "' was given.")
         
-        args = {"grid": grid}
-        self._constraints[grid].append((constr, get_meta(), args))
+        args = {"grid": grid, "include_last": include_last, "include_first": include_first}
+        self._constraints[grid].append((constr, get_meta(meta), args))
 
     def at_t0(self, expr):
         """Evaluate a signal at the start of the horizon
@@ -523,8 +578,9 @@ class Stage:
         """
         from copy import deepcopy
         self._set_transcribed(False)
+        template = self._method
         self._method = deepcopy(method)
-        self._method.register(self)
+        self._method.inherit(template)
 
     @property
     def objective(self):
@@ -549,7 +605,8 @@ class Stage:
 
     @property
     def p(self):
-        return vvcat(self.parameters['']+self.parameters['control'])
+        arg = self.parameters['']+self.parameters['control']
+        return MX(0, 1) if len(arg)==0 else vvcat(arg)
 
     @property
     def v(self):
@@ -571,6 +628,19 @@ class Stage:
     def np(self):
         return self.p.numel()
 
+    @property
+    def gist(self):
+        """Obtain an expression packing all information needed to obtain value/sample
+
+        The composition of this array may vary between rockit versions
+
+        Returns
+        -------
+        :obj:`~casadi.MX` column vector
+
+        """
+        return self.master.gist
+
     def is_signal(self, expr):
         """Does the expression represent a signal (does it depend on time)?
 
@@ -580,7 +650,7 @@ class Stage:
 
         """
  
-        return depends_on(expr, vertcat(self.x, self.u, self.t, vcat(self.variables['control']+self.variables['states']), vvcat(self._inf_der.keys())))
+        return depends_on(expr, vertcat(self.x, self.u, self.z, self.t, vcat(self.variables['control']+self.variables['states']), vvcat(self._inf_der.keys())))
 
     def _create_placeholder_expr(self, expr, callback_name):
         r = MX.sym("r_" + callback_name, MX(expr).sparsity())
@@ -603,6 +673,14 @@ class Stage:
         alg = veccat(*self._alg)
         return Function('ode', [self.x, self.u, self.z, vertcat(self.p, self.v), self.t], [ode, alg, quad], ["x", "u", "z", "p", "t"], ["ode","alg","quad"])
 
+    # Internal methods
+    def _diffeq(self):
+        next = veccat(*[self._state_next[k] for k in self.states])
+        quad = veccat(*[self._state_next[k] for k in self.qstates])
+
+        dt = MX(1,1)
+        return Function('ode', [self.x, self.u, vertcat(self.p, self.v), self.t, dt], [next, MX(), quad, MX(0, 1), MX()], ["x0", "u", "p", "t0", "DT"], ["xf","poly_coeff","qf","zf","poly_coeff_z"])
+
     def _expr_apply(self, expr, **kwargs):
         """
         Substitute placeholder symbols with actual decision variables,
@@ -676,13 +754,16 @@ class Stage:
         else:
             return False
 
-    def _transcribe_recurse(self):
-        opti = self.master.opti
+    def _transcribe_recurse(self, pass_nr=1, **kwargs):
         if self._method is not None:
-            self._method.transcribe(self, opti)
+            if self is self.master:
+                self._method.main_transcribe(self, pass_nr=pass_nr, **kwargs)
+            self._method.transcribe(self, pass_nr=pass_nr, **kwargs)
+        else:
+            print("master",self)
 
         for s in self._stages:
-            s._transcribe_recurse()
+            s._transcribe_recurse(pass_nr=pass_nr, **kwargs)
 
     def _placeholders_transcribe_recurse(self, placeholders):
         if self._method is not None:
@@ -699,9 +780,9 @@ class Stage:
         subst_from = list(self._placeholders.keys())
         subst_to = []
         for k in self._placeholders.keys():
-            if k is self.T:  # T and t0 already have new placeholder symbols
+            if is_equal(k, self.T):  # T and t0 already have new placeholder symbols
                 subst_to.append(ret.T)
-            elif k is self.t0:
+            elif is_equal(k, self.t0):
                 subst_to.append(ret.t0)
             else:
                 subst_to.append(MX.sym(k.name(), k.sparsity()))
@@ -711,11 +792,15 @@ class Stage:
 
         ret.states = copy(self.states)
         ret.controls = copy(self.controls)
+        ret.algebraics = copy(self.algebraics)
         ret.parameters = deepcopy(self.parameters)
         ret.variables = deepcopy(self.variables)
 
+        ret._offsets = deepcopy(self._offsets)
         ret._param_vals = copy(self._param_vals)
         ret._state_der = copy(self._state_der)
+        ret._alg = copy(self._alg)
+        ret._state_next = copy(self._state_next)
         constr_types = self._constraints.keys()
         orig = []
         for k in constr_types:
@@ -732,14 +817,14 @@ class Stage:
             ret._constraints[k] = list(zip(r, [merge_meta(m, get_meta()) for _, m, _ in v], [d for _, _, d in v]))
             r = r[len(v):]
 
-        ret._initial = dict(zip(res[n_constr+1:], self._initial.values()))
+        ret._initial = HashOrderedDict(zip(res[n_constr+1:], self._initial.values()))
         ret._check_free_time()
 
         if "T" not in kwargs:
             ret._T = copy(self._T)
         if "t0" not in kwargs:
             ret._t0 = copy(self._t0)
-        ret.t = self.t
+        ret._t = self.t
         ret._method = deepcopy(self._method)
 
         ret._is_transcribed = False
@@ -815,7 +900,7 @@ class Stage:
 
         return placeholders(time), placeholders(res)
 
-    def _grid_control(self, stage, expr, grid, include_first=True, include_last=True):
+    def _grid_control(self, stage, expr, grid, include_first=True, include_last=True, transpose=False):
         """Evaluate expression at (N + 1) control points."""
         sub_expr = []
         ks = list(range(1, stage._method.N))
@@ -824,8 +909,13 @@ class Stage:
         if include_last:
             ks = ks+[-1]
         for k in ks:
-            sub_expr.append(stage._method.eval_at_control(stage, expr, k))
-        res = hcat(sub_expr)
+            try:
+                r = stage._method.eval_at_control(stage, expr, k)
+            except IndexError as e:
+                r = DM.nan(MX(expr).shape)
+            sub_expr.append(r)
+        cat = vcat if transpose else hcat
+        res = cat(sub_expr)
         time = stage._method.control_grid
         return time, res
 
@@ -886,7 +976,7 @@ class Stage:
                     z = mtimes(coeff_z,tpower_z)
                 else:
                     z = nan
-                sub_expr.append(expr_f(mtimes(coeff,tpower), z, stage._method.U[k]))
+                sub_expr.append(stage._method.eval_at_integrator(stage, expr_f(mtimes(coeff,tpower), z, stage._method.U[k]), k, l))
                 t0+=dt
 
         ts = tlocal[-1,:]
@@ -898,7 +988,7 @@ class Stage:
         else:
             z = nan
 
-        sub_expr.append(expr_f(mtimes(stage._method.poly_coeff[-1],tpower), z, stage._method.U[-1]))
+        sub_expr.append(stage._method.eval_at_integrator(stage, expr_f(mtimes(stage._method.poly_coeff[-1],tpower), z, stage._method.U[-1]), k, l))
 
         return vcat(total_time), hcat(sub_expr)
 
@@ -914,6 +1004,10 @@ class Stage:
         placeholders = self.master.placeholders_transcribed
         return placeholders(self._method.eval(self, expr))
 
+    def discrete_system(self):
+        """Hack"""
+        return self._method.discrete_system(self)
+
     def sampler(self, *args):
         """Returns a function that samples given expressions
 
@@ -1011,7 +1105,7 @@ class Stage:
                 :obj:`np.array`
 
                 """
-                tdim = None if isinstance(t, float) or len(t.shape)==0 else DM(t).numel()
+                tdim = None if isinstance(t, float) or isinstance(t, int) or len(t.shape)==0 else DM(t).numel()
                 t = DM(t)
                 if t.is_column(): t = t.T
                 res = f.call([gist, t])
